# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Module houses class that wraps data (block partition) and its metadata."""

import pandas

from modin.data_management.utils import length_fn_pandas, width_fn_pandas
from modin.engines.base.frame.partition import PandasFramePartition

from distributed.client import default_client
from distributed import Future
from distributed.utils import get_ip
from dask.distributed import wait

from modin.pandas.indexing import compute_sliced_len


class PandasOnDaskFramePartition(PandasFramePartition):
    """
    The class implements the interface in ``PandasFramePartition``.

    Parameters
    ----------
    future : distributed.Future
        A reference to pandas DataFrame that need to be wrapped with this class.
    length : distributed.Future or int, optional
        Length or reference to it of wrapped pandas DataFrame.
    width : distributed.Future or int, optional
        Width or reference to it of wrapped pandas DataFrame.
    ip : distributed.Future or str, optional
        Node IP address or reference to it that holds wrapped pandas DataFrame.
    call_queue : list, optional
        Call queue that needs to be executed on wrapped pandas DataFrame.
    """

    def __init__(self, future, length=None, width=None, ip=None, call_queue=None):
        self.future = future
        if call_queue is None:
            call_queue = []
        self.call_queue = call_queue
        self._length_cache = length
        self._width_cache = width
        self._ip_cache = ip

    def get(self):
        """
        Get the object wrapped by this partition out of the distributed memory.

        Returns
        -------
        pandas.DataFrame
            The object from the distributed memory.
        """
        self.drain_call_queue()
        # blocking operation
        if isinstance(self.future, pandas.DataFrame):
            return self.future
        return self.future.result()

    def apply(self, func, *args, **kwargs):
        """
        Apply a function to the object wrapped by this partition.

        Parameters
        ----------
        func : callable
            A function to apply.
        *args : iterable
            Additional positional arguments to be passed in `func`.
        **kwargs : dict
            Additional keyword arguments to be passed in `func`.

        Returns
        -------
        PandasOnDaskFramePartition
            A new ``PandasOnDaskFramePartition`` object.

        Notes
        -----
        The keyword arguments are sent as a dictionary.
        """
        client = default_client()
        call_queue = self.call_queue + [[func, args, kwargs]]
        if len(call_queue) > 1:
            future = client.submit(
                apply_list_of_funcs, call_queue, self.future, pure=False
            )
        else:
            # We handle `len(call_queue) == 1` in a different way because
            # this improves performance a bit.
            func, args, kwargs = call_queue[0]
            future = client.submit(apply_func, self.future, func, *args, **kwargs)
        futures = [
            client.submit(lambda l, i: l[i], future, i, pure=False) for i in range(2)
        ]
        return PandasOnDaskFramePartition(futures[0], ip=futures[1])

    def add_to_apply_calls(self, func, *args, **kwargs):
        """
        Add a function to the call queue.

        Parameters
        ----------
        func : callable
            Function to be added to the call queue.
        *args : iterable
            Additional positional arguments to be passed in `func`.
        **kwargs : dict
            Additional keyword arguments to be passed in `func`.

        Returns
        -------
        PandasOnDaskFramePartition
            A new ``PandasOnDaskFramePartition`` object.

        Notes
        -----
        The keyword arguments are sent as a dictionary.
        """
        return PandasOnDaskFramePartition(
            self.future, call_queue=self.call_queue + [[func, args, kwargs]]
        )

    def drain_call_queue(self):
        """Execute all operations stored in the call queue on the object wrapped by this partition."""
        if len(self.call_queue) == 0:
            return
        call_queue = self.call_queue
        client = default_client()
        if len(call_queue) > 1:
            future = client.submit(
                apply_list_of_funcs, call_queue, self.future, pure=False
            )
        else:
            # We handle `len(call_queue) == 1` in a different way because
            # this improves performance a bit.
            func, args, kwargs = call_queue[0]
            future = client.submit(apply_func, self.future, func, *args, **kwargs)
        futures = [
            client.submit(lambda l, i: l[i], future, i, pure=False) for i in range(2)
        ]
        self.future = futures[0]
        self._ip_cache = futures[1]
        self.call_queue = []

    def wait(self):
        """Wait completing computations on the object wrapped by the partition."""
        self.drain_call_queue()
        wait(self.future)

    def mask(self, row_indices, col_indices):
        """
        Lazily create a mask that extracts the indices provided.

        Parameters
        ----------
        row_indices : list-like, slice or label
            The indices for the rows to extract.
        col_indices : list-like, slice or label
            The indices for the columns to extract.

        Returns
        -------
        PandasOnDaskFramePartition
            A new ``PandasOnDaskFramePartition`` object.
        """
        new_obj = super().mask(row_indices, col_indices)
        client = default_client()
        if isinstance(row_indices, slice) and isinstance(self._length_cache, Future):
            new_obj._length_cache = client.submit(
                compute_sliced_len, row_indices, self._length_cache
            )
        if isinstance(col_indices, slice) and isinstance(self._width_cache, Future):
            new_obj._width_cache = client.submit(
                compute_sliced_len, col_indices, self._width_cache
            )
        return new_obj

    def __copy__(self):
        """
        Create a copy of this partition.

        Returns
        -------
        PandasOnDaskFramePartition
            A copy of this partition.
        """
        return PandasOnDaskFramePartition(
            self.future,
            length=self._length_cache,
            width=self._width_cache,
            ip=self._ip_cache,
            call_queue=self.call_queue,
        )

    def to_pandas(self):
        """
        Convert the object wrapped by this partition to a pandas DataFrame.

        Returns
        -------
        pandas.DataFrame
        """
        dataframe = self.get()
        assert type(dataframe) is pandas.DataFrame or type(dataframe) is pandas.Series

        return dataframe

    def to_numpy(self, **kwargs):
        """
        Convert the object wrapped by this partition to a NumPy array.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments to be passed in ``to_numpy``.

        Returns
        -------
        np.ndarray.
        """
        return self.apply(lambda df, **kwargs: df.to_numpy(**kwargs)).get()

    @classmethod
    def put(cls, obj):
        """
        Put an object into distributed memory and wrap it with partition object.

        Parameters
        ----------
        obj : any
            An object to be put.

        Returns
        -------
        PandasOnDaskFramePartition
            A new ``PandasOnDaskFramePartition`` object.
        """
        client = default_client()
        return cls(client.scatter(obj, hash=False))

    @classmethod
    def preprocess_func(cls, func):
        """
        Preprocess a function before an ``apply`` call.

        Parameters
        ----------
        func : callable
            The function to preprocess.

        Returns
        -------
        callable
            An object that can be accepted by ``apply``.
        """
        return default_client().scatter(func, hash=False, broadcast=True)

    @classmethod
    def _length_extraction_fn(cls):
        """
        Return the function that computes the length of the object wrapped by this partition.

        Returns
        -------
        callable
            The function that computes the length of the object wrapped by this partition.
        """
        return length_fn_pandas

    @classmethod
    def _width_extraction_fn(cls):
        """
        Return the function that computes the width of the object wrapped by this partition.

        Returns
        -------
        callable
            The function that computes the width of the object wrapped by this partition.
        """
        return width_fn_pandas

    _length_cache = None
    _width_cache = None

    def length(self):
        """
        Get the length of the object wrapped by this partition.

        Returns
        -------
        int
            The length of the object.
        """
        if self._length_cache is None:
            self._length_cache = self.apply(lambda df: len(df)).future
        if isinstance(self._length_cache, Future):
            self._length_cache = self._length_cache.result()
        return self._length_cache

    def width(self):
        """
        Get the width of the object wrapped by the partition.

        Returns
        -------
        int
            The width of the object.
        """
        if self._width_cache is None:
            self._width_cache = self.apply(lambda df: len(df.columns)).future
        if isinstance(self._width_cache, Future):
            self._width_cache = self._width_cache.result()
        return self._width_cache

    def ip(self):
        """
        Get the node IP address of the object wrapped by this partition.

        Returns
        -------
        str
            IP address of the node that holds the data.
        """
        if self._ip_cache is None:
            self._ip_cache = self.apply(lambda df: df)._ip_cache
        if isinstance(self._ip_cache, Future):
            self._ip_cache = self._ip_cache.result()
        return self._ip_cache

    @classmethod
    def empty(cls):
        """
        Create a new partition that wraps an empty pandas DataFrame.

        Returns
        -------
        PandasOnDaskFramePartition
            A new ``PandasOnDaskFramePartition`` object.
        """
        return cls(pandas.DataFrame(), 0, 0)


def apply_func(partition, func, *args, **kwargs):
    """
    Execute a function on the partition in a worker process.

    Parameters
    ----------
    partition : pandas.DataFrame
        A pandas DataFrame the function needs to be executed on.
    func : callable
        Function that needs to be executed on `partition`.
    *args
        Additional positional arguments to be passed in `func`.
    **kwargs
        Additional keyword arguments to be passed in `func`.

    Returns
    -------
    pandas.DataFrame
        The resulting pandas DataFrame.
    str
        The node IP address of the worker process.
    """
    result = func(partition, *args, **kwargs)
    return result, get_ip()


def apply_list_of_funcs(funcs, partition):
    """
    Execute all operations stored in the call queue on the partition in a worker process.

    Parameters
    ----------
    funcs : list
        A call queue that needs to be executed on the partition.
    partition : pandas.DataFrame
        A pandas DataFrame the call queue needs to be executed on.

    Returns
    -------
    pandas.DataFrame
        The resulting pandas DataFrame.
    str
        The node IP address of the worker process.
    """
    for func, args, kwargs in funcs:
        partition = func(partition, *args, **kwargs)
    return partition, get_ip()



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import pandas
import pytest
import modin.pandas as pd
import modin.spreadsheet as mss
import numpy as np
from modin_spreadsheet import SpreadsheetWidget


def get_test_data():
    return {
        "A": 1.0,
        "B": pd.Timestamp("20130102"),
        "C": pd.Series(1, index=list(range(4)), dtype="float32"),
        "D": np.array([5, 2, 3, 1], dtype="int32"),
        "E": pd.Categorical(["test", "train", "foo", "bar"]),
        "F": ["foo", "bar", "buzz", "fox"],
    }


def test_from_dataframe():
    data = get_test_data()
    modin_df = pd.DataFrame(data)
    pandas_df = pandas.DataFrame(data)

    modin_result = mss.from_dataframe(modin_df)
    assert isinstance(modin_result, SpreadsheetWidget)

    with pytest.raises(TypeError):
        mss.from_dataframe(pandas_df)

    # Check parameters don't error
    def can_edit_row(row):
        return row["D"] > 2

    modin_result = mss.from_dataframe(
        modin_df,
        show_toolbar=True,
        show_history=True,
        precision=1,
        grid_options={"forceFitColumns": False, "filterable": False},
        column_options={"D": {"editable": True}},
        column_definitions={"editable": False},
        row_edit_callback=can_edit_row,
    )
    assert isinstance(modin_result, SpreadsheetWidget)


def test_to_dataframe():
    data = get_test_data()
    modin_df = pd.DataFrame(data)
    pandas_df = pandas.DataFrame(data)

    spreadsheet = mss.from_dataframe(modin_df)
    modin_result = mss.to_dataframe(spreadsheet)

    assert modin_result.equals(modin_df)

    with pytest.raises(TypeError):
        mss.to_dataframe("Not a SpreadsheetWidget")
    with pytest.raises(TypeError):
        mss.to_dataframe(pandas_df)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""The module defines interface for an axis partition with PyArrow backend and Ray engine."""

from modin.engines.base.frame.axis_partition import BaseFrameAxisPartition
from .partition import PyarrowOnRayFramePartition

import ray
import pyarrow


class PyarrowOnRayFrameAxisPartition(BaseFrameAxisPartition):
    """
    Class defines axis partition interface with PyArrow backend and Ray engine.

    Inherits functionality from ``BaseFrameAxisPartition`` class.

    Parameters
    ----------
    list_of_blocks : list
        List with partition objects to create common axis partition for.
    """

    def __init__(self, list_of_blocks):
        # Unwrap from PandasFramePartition object for ease of use
        self.list_of_blocks = [obj.oid for obj in list_of_blocks]

    def apply(self, func, num_splits=None, other_axis_partition=None, **kwargs):
        """
        Apply func to the object in the Plasma store.

        Parameters
        ----------
        func : callable or ray.ObjectRef
            The function to apply.
        num_splits : int, optional
            The number of times to split the resulting object.
        other_axis_partition : PyarrowOnRayFrameAxisPartition, optional
            Another ``PyarrowOnRayFrameAxisPartition`` object to apply to
            `func` with this one.
        **kwargs : dict
            Additional keyward arguments to pass with `func`.

        Returns
        -------
        list
            List with ``PyarrowOnRayFramePartition`` objects.

        Notes
        -----
        See notes in Parent class about this method.
        """
        if num_splits is None:
            num_splits = len(self.list_of_blocks)

        if other_axis_partition is not None:
            return [
                PyarrowOnRayFramePartition(obj)
                for obj in deploy_ray_func_between_two_axis_partitions.options(
                    num_returns=num_splits
                ).remote(
                    self.axis,
                    func,
                    num_splits,
                    len(self.list_of_blocks),
                    kwargs,
                    *(self.list_of_blocks + other_axis_partition.list_of_blocks),
                )
            ]

        args = [self.axis, func, num_splits, kwargs]
        args.extend(self.list_of_blocks)
        return [
            PyarrowOnRayFramePartition(obj)
            for obj in deploy_ray_axis_func.options(num_returns=num_splits).remote(
                *args
            )
        ]

    def shuffle(self, func, num_splits=None, **kwargs):
        """
        Shuffle the order of the data in this axis based on the `func`.

        Parameters
        ----------
        func : callable
            The function to apply before splitting.
        num_splits : int, optional
            The number of times to split the resulting object.
        **kwargs : dict
            Additional keywords arguments to be passed in `func`.

        Returns
        -------
        list
            List with ``PyarrowOnRayFramePartition`` objects.

        Notes
        -----
        Method extends ``BaseFrameAxisPartition.shuffle``.
        """
        if num_splits is None:
            num_splits = len(self.list_of_blocks)

        args = [self.axis, func, num_splits, kwargs]
        args.extend(self.list_of_blocks)
        return [
            PyarrowOnRayFramePartition(obj)
            for obj in deploy_ray_axis_func.options(num_returns=num_splits).remote(
                *args
            )
        ]


class PyarrowOnRayFrameColumnPartition(PyarrowOnRayFrameAxisPartition):
    """
    The column partition implementation for PyArrow backend and Ray engine.

    All of the implementation for this class is in the ``PyarrowOnRayFrameAxisPartition``
    parent class, and this class defines the axis to perform the computation over.

    Parameters
    ----------
    list_of_blocks : list
        List with partition objects to create common axis partition.
    """

    axis = 0


class PyarrowOnRayFrameRowPartition(PyarrowOnRayFrameAxisPartition):
    """
    The row partition implementation for PyArrow backend and Ray engine.

    All of the implementation for this class is in the ``PyarrowOnRayFrameAxisPartition``
    parent class, and this class defines the axis to perform the computation over.

    Parameters
    ----------
    list_of_blocks : list
        List with partition objects to create common axis partition.
    """

    axis = 1


def concat_arrow_table_partitions(axis, partitions):
    """
    Concatenate given `partitions` in a single table.

    Parameters
    ----------
    axis : {0, 1}
        The axis to concatenate over.
    partitions : array-like
        Array with partitions for concatenating.

    Returns
    -------
    pyarrow.Table
        ``pyarrow.Table`` constructed from the passed partitions.
    """
    if axis == 0:
        table = pyarrow.Table.from_batches(
            [part.to_batches(part.num_rows)[0] for part in partitions]
        )
    else:
        table = partitions[0].drop([partitions[0].columns[-1].name])
        for obj in partitions[1:]:
            i = 0
            for col in obj.itercolumns():
                if i < obj.num_columns - 1:
                    table = table.append_column(col)
                i += 1
        table = table.append_column(partitions[0].columns[-1])
    return table


def split_arrow_table_result(axis, result, num_partitions, num_splits, metadata):
    """
    Split ``pyarrow.Table`` according to the passed parameters.

    Parameters
    ----------
    axis : {0, 1}
        The axis to perform the function along.
    result : pyarrow.Table
        Resulting table to split.
    num_partitions : int
        Number of partitions that `result` was constructed from.
    num_splits : int
        The number of splits to return.
    metadata : dict
        Dictionary with ``pyarrow.Table`` metadata.

    Returns
    -------
    list
        List of PyArrow Tables.
    """
    chunksize = (
        num_splits // num_partitions
        if num_splits % num_partitions == 0
        else num_splits // num_partitions + 1
    )
    if axis == 0:
        return [
            pyarrow.Table.from_batches([part]) for part in result.to_batches(chunksize)
        ]
    else:
        return [
            result.drop(
                [
                    result.columns[i].name
                    for i in range(result.num_columns)
                    if i >= n * chunksize or i < (n - 1) * chunksize
                ]
            )
            .append_column(result.columns[-1])
            .replace_schema_metadata(metadata=metadata)
            for n in range(1, num_splits)
        ] + [
            result.drop(
                [
                    result.columns[i].name
                    for i in range(result.num_columns)
                    if i < (num_splits - 1) * chunksize
                ]
            ).replace_schema_metadata(metadata=metadata)
        ]


@ray.remote
def deploy_ray_axis_func(axis, func, num_splits, kwargs, *partitions):
    """
    Deploy a function along a full axis in Ray.

    Parameters
    ----------
    axis : {0, 1}
        The axis to perform the function along.
    func : callable
        The function to deploy.
    num_splits : int
        The number of splits to return.
    kwargs : dict
        A dictionary of keyword arguments.
    *partitions : array-like
        All partitions that make up the full axis (row or column).

    Returns
    -------
    list
        List of PyArrow Tables.
    """
    table = concat_arrow_table_partitions(axis, partitions)
    try:
        result = func(table, **kwargs)
    except Exception:
        result = pyarrow.Table.from_pandas(func(table.to_pandas(), **kwargs))
    return split_arrow_table_result(
        axis, result, len(partitions), num_splits, table.schema.metadata
    )


@ray.remote
def deploy_ray_func_between_two_axis_partitions(
    axis, func, num_splits, len_of_left, kwargs, *partitions
):
    """
    Deploy a function along a full axis between two data sets in Ray.

    Parameters
    ----------
    axis : {0, 1}
        The axis to perform the function along.
    func : callable
        The function to deploy.
    num_splits : int
        The number of splits to return.
    len_of_left : int
        The number of values in `partitions` that belong to the left data set.
    kwargs : dict
        A dictionary of keyword arguments.
    *partitions : array-like
        All partitions that make up the full axis (row or column)
        for both data sets.

    Returns
    -------
    list
        List of PyArrow Tables.
    """
    lt_table = concat_arrow_table_partitions(axis, partitions[:len_of_left])
    rt_table = concat_arrow_table_partitions(axis, partitions[len_of_left:])
    try:
        result = func(lt_table, rt_table, **kwargs)
    except Exception:
        lt_frame = lt_table.from_pandas()
        rt_frame = rt_table.from_pandas()
        result = pyarrow.Table.from_pandas(func(lt_frame, rt_frame, **kwargs))
    return split_arrow_table_result(
        axis, result, len(result.num_rows), num_splits, result.schema.metadata
    )


@ray.remote
def deploy_ray_shuffle_func(axis, func, num_splits, kwargs, *partitions):
    """
    Deploy shuffle function that defines the order of the data in this axis partition.

    Parameters
    ----------
    axis : {0, 1}
        The axis to perform the function along.
    func : callable
        The function to deploy.
    num_splits : int
        The number of splits to return.
    kwargs : dict
        A dictionary of keyword arguments.
    *partitions : array-like
        All partitions that make up the full axis (row or column).

    Notes
    -----
    Function is deprecated.
    """



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import os
from os.path import dirname, abspath


# This is the python file root directory (modin/modin)
rootdir = dirname(abspath(__file__))
exclude_files = ["_version.py"]

with open("{}{}".format(dirname(rootdir), "/LICENSE_HEADER"), "r") as f:
    # Lines to check each line individually
    header_lines = f.readlines()

for subdir, dirs, files in os.walk(rootdir):
    for file in files:
        filepath = os.path.join(subdir, file)
        if file.endswith(".py") and file not in exclude_files:
            with open(filepath, "r+") as f:
                # Lines for line by line comparison
                py_file_lines = f.readlines()
                if len(py_file_lines):
                    first_non_comment_idx = next(
                        i
                        for i in range(len(py_file_lines))
                        if not py_file_lines[i].startswith("# ")
                    )
                    py_file_lines = py_file_lines[first_non_comment_idx:]
                new_contents = "".join(header_lines + py_file_lines)
                f.seek(0)
                f.write(new_contents)
                f.truncate()
